from ta.base_indicator import BaseIndicator


class VolumeChangeMonitor(BaseIndicator):
    def __init__(self, market, interval):
        super(VolumeChangeMonitor, self).__init__(market, interval, 2)
        self.__previous_volume = 0
        self.value = None

    def next_calculation(self, candle):
        """get latest N candles from market, do calculation, write results to db"""
        new_volume = candle[5]
        if self.__previous_volume is not 0:
            self.value = round(100 * ((new_volume - self.__previous_volume) / self.__previous_volume), 2)  # calculate change in volume in percentage terms
        self.__previous_volume = new_volume




from core.markets.market import Market
from core.database import ohlcv_functions
from core.markets import position
from core.markets import order
import logging

logger = logging.getLogger(__name__)

long_positions = 0


class MarketSimulator(Market):
    """Wrapper for market that allows simulating simple buys and sells"""
    def __init__(self, exchange, base_currency, quote_currency, quote_currency_balance, strategy):
        super().__init__(exchange, base_currency, quote_currency, strategy)
        self.starting_balance = quote_currency_balance
        self.quote_balance = quote_currency_balance
        self.base_balance = 0
        self.simulating = False

    def limit_buy(self, quantity, price):
        if self.quote_balance >= quantity * price:
            self.quote_balance = self.quote_balance - quantity * price
            self.base_balance = self.base_balance + quantity
            order.write_order_to_db(self.exchange.id, self.analysis_pair, "buy", quantity, price, "simulated")
            logger.info("Executed buy simulation of " + str(quantity) + " " + self.base_currency + " for " + str(price) + " " + self.quote_currency)
            logger.info(self.quote_currency + " balance: " + str(self.quote_balance))
            logger.info(self.base_currency + " balance: " + str(self.base_balance))
        else:
            logger.info("Insufficient balance for simulation buy")

    def limit_sell(self, quantity, price):
        if self.base_balance >= quantity:
            self.base_balance = self.base_balance - quantity
            self.quote_balance = self.quote_balance + quantity * price
            order.write_order_to_db(self.exchange.id, self.analysis_pair, "sell", quantity, price, "simulated")
            logger.info("Executed sell simulation of " + str(quantity) + " " + self.base_currency + " for " + str(price) + " " + self.quote_currency)
            logger.info(self.quote_currency + " balance: " + str(self.quote_balance))
            logger.info(self.base_currency + " balance: " + str(self.base_balance))
        else:
            logger.info("Insufficient balance for simulation sell")

    def market_buy(self, quantity):
        if self.quote_balance >= quantity * self.get_ask_price():
            self.quote_balance = self.quote_balance - quantity * self.get_ask_price()
            self.base_balance = self.base_balance + quantity
            logger.info("Executed buy simulation of " + str(quantity) + " " + self.base_currency + " for " + str(self.get_ask_price()) + " " + self.quote_currency)
            logger.info(self.quote_currency + " balance: " + str(self.quote_balance))
            logger.info(self.base_currency + " balance: " + str(self.base_balance))
        else:
            logger.info("Insufficient balance for simulation buy")

    def market_sell(self, quantity):
        if self.base_balance >= quantity:
            self.base_balance = self.base_balance - quantity
            self.quote_balance = self.quote_balance + quantity * self.get_bid_price()
            logger.info("Executed sell simulation of " + str(quantity) + " " + self.base_currency + " for " + str(self.get_bid_price()) + " " + self.quote_currency)
            logger.info(self.quote_currency + " balance: " + str(self.quote_balance))
            logger.info(self.base_currency + " balance: " + str(self.base_balance))
        else:
            logger.info("Insufficient balance for simulation sell")

    def get_ask_price(self):
        """Get ask price for simulation"""
        if not self.simulating:
            """if operating on live data, use actual ask"""
            return self.exchange.fetchTicker(self.analysis_pair)['ask']
        else:
            """if operating on historical data, use close"""
            return self.latest_candle['5m'][4]

    def get_bid_price(self):
        if not self.simulating:
            """if operating on live data, use actual ask"""
            return self.exchange.fetchTicker(self.analysis_pair)['bid']
        else:
            """if operating on historical data, use close"""
            return self.latest_candle['5m'][4]

    def get_wallet_balance(self):
        return self.quote_balance


def open_long_position_simulation(market, amount, price, fixed_stoploss, trailing_stoploss_percent, profit_target_percent):
    """Create simulated long position"""
    logger.info("Opening simulated long position")
    position = LongPositionSimulator(market, amount, price, fixed_stoploss, trailing_stoploss_percent, profit_target_percent)
    position.open()
    return position


def open_short_position_simulation(market, amount, price):
    """Create simulated short position"""
    logger.info("Opening simulated short position")
    position = ShortPositionSimulator(market, amount, price)
    position.open()
    return position


class LongPositionSimulator(position.LongPosition):
    """Simulated long position. Overrides the functionality of creating an actual order to use the MarketSimulators balance and calculations"""
    def __init__(self, market, amount, price, fixed_stoploss, trailing_stoploss_percent, profit_target_percent):
        super().__init__(market, amount, price, fixed_stoploss, trailing_stoploss_percent, profit_target_percent)

    def liquidate_position(self):
        """Will use this method to actually create the order that liquidates the position"""
        logger.info("Closing simulated long position")
        open_short_position_simulation(self.market, self.amount, self.market.latest_candle['5m'][3])
        self.is_open = False

    def open(self):
        self.market.limit_buy(self.amount, self.price)
        self.is_open = True

    def update(self):
        """Use this method to trigger position to check if profit target has been met, and re-set trailiing stop loss"""
        logger.info("UPDATING LONG POSITION")
        if self.market.latest_candle['5m'][3] < self.trailing_stoploss or\
            self.market.latest_candle['5m'][3] < self.fixed_stoploss or\
            self.market.latest_candle['5m'][3] >= self.profit_target:  # check price against last calculated trailing stoploss
                self.liquidate_position()
        # re-calculate trailing stoploss
        self.trailing_stoploss = self.calculate_trailing_stoploss()


class ShortPositionSimulator(position.ShortPosition):
    """Simulated short position. Overrides the functionality of creating an actual order to use the MarketSimulators balance and calculations"""
    def __init__(self, market, amount, price):
        super().__init__(market, amount, price)

    def open(self):
        self.market.limit_sell(self.amount, self.price)


from collections import defaultdict
from threading import Thread
from queue import Queue
from core.database import ohlcv_functions
from core.markets import ticker
from ccxt import BaseError
import ccxt
import logging
import time
from pubsub import pub
from threading import Lock

lock = Lock()
logger = logging.getLogger(__name__)

class MarketWatcher:
    """Active object that subscribes to a ticker of a specific interval and keeps track of OHLCV data
     A market watcher is instantiated with a trading pair (base, quote) and an interval
     It then subscribes to the ticker of that interval and calls for candles each time period
     It is responsible for syncing data with the DB and adding new candles
     Strategies that subscribe to the ticker will be given the new candles"""
    def __init__(self, exchange, base_currency, quote_currency, interval):
        exchange = getattr(ccxt, exchange)
        ticker.subscribe(self.tick, interval)
        self.analysis_pair = '{}/{}'.format(base_currency, quote_currency)
        self.exchange = exchange()
        self.interval = interval
        self.base_currency = base_currency
        self.quote_currency = quote_currency
        self.topic = self.exchange.id + self.analysis_pair + self.interval
        self.__thread = Thread(target=self.__run)  # create thread for listener
        self._jobs = Queue()  # create job queue
        self.__running = False
        self.historical_synced = False
        self.latest_candle = None
        self.PairID = ohlcv_functions.write_trade_pairs_to_db(self.exchange.id, self.base_currency, self.quote_currency, self.interval)
        self.__thread.start()

    def __run(self):
        """Start listener queue waiting for ticks"""
        self.__running = True
        self.sync_historical()
        while self.__running:
            if not self._jobs.empty():
                job = self._jobs.get()
                try:
                    job()
                except Exception as e:
                    print(e)
                    logger.error(job.__name__ + " threw error:\n" + str(e))

    def stop(self):
        """Stop listener queue"""
        self.__running = False

    def tick(self):
        """Queue a pull of the latest candle"""
        if self.historical_synced:
            self._jobs.put(lambda: self.__pull_latest_candle(self.interval))

    def sync_historical(self):
        """Queue loading of historical candles"""
        self._jobs.put(lambda: self.__sync_historical())

    def get_historical_candles(self):
        data = ohlcv_functions.get_all_candles(self.PairID)
        return data

    def __sync_historical(self):
        """Load all missing historical candles to database"""
        logger.info('Syncing market candles with DB...')
        latest_db_candle = ohlcv_functions.get_latest_candle(self.exchange.id, self.analysis_pair, self.interval)
        data = self.exchange.fetch_ohlcv(self.analysis_pair, self.interval)
        if latest_db_candle is None:
            logger.info("No historical data for market, adding all available OHLCV data")
            for entry in data:
                ohlcv_functions.insert_data_into_ohlcv_table(self.exchange.id, self.analysis_pair, self.interval, entry, self.PairID)
                print('Writing candle ' + str(entry[0]) + ' to database')
        else:
            for entry in data:
                if not latest_db_candle[10] >= entry[0]:
                    ohlcv_functions.insert_data_into_ohlcv_table(self.exchange.id, self.analysis_pair, self.interval, entry, self.PairID)
                    print('Writing missing candle ' + str(entry[0]) + ' to database')
        self.historical_synced = True
        pub.sendMessage(self.topic + "historical")
        logger.info('Market data has been synced.')

    def __pull_latest_candle(self, interval):
        """Initiate a pull of the latest candle, making sure not to pull a duplicate candle"""
        logger.info("Getting latest candle for " + self.exchange.id + " " + self.analysis_pair + " " + interval)
        print("Getting latest candle")
        try:
            latest_data = self.exchange.fetch_ohlcv(self.analysis_pair, interval)[-1]
            while latest_data == self.latest_candle:
                print("retrying candle update")
                logger.info('Candle already contained in DB, retrying...')
                time.sleep(self.exchange.rateLimit * 2 / 1000)
                latest_data = self.exchange.fetch_ohlcv(self.analysis_pair, interval)[-1]
            ohlcv_functions.insert_data_into_ohlcv_table(self.exchange.id, self.analysis_pair, interval, latest_data, self.PairID)
        except Exception as e:
            print(e)
            logger.info("Timeout pulling latest candle, trying again")
            self.__pull_latest_candle(interval)
            return
        self.latest_candle = latest_data
        pub.sendMessage(self.topic, candle=self.latest_candle)
        print("Sent message to " + self.topic)


lookup_list = defaultdict(MarketWatcher)


def get_market_watcher(exchange_id, base, quote, interval):
    """Return or create market watcher for the given analysis market"""
    topic = str(exchange_id + base + "/" + quote + interval)
    if topic not in lookup_list:
        lookup_list[topic] = MarketWatcher(exchange_id, base, quote, interval)
    return lookup_list[topic]


def subscribe_historical(exchange_id, base, quote, interval, callable):
    """Subscribe to a notification that is sent when historical data is loaded for the market given"""
    topic = str(exchange_id + base + "/" + quote + interval + "historical")
    pub.subscribe(callable, topic)


def subscribe(exchange_id, base, quote, interval, callable):
    """
    Enroll strategy to recieve new candles from a market
    :param exchange_id: string representing exchange i.e. 'bittrex'
    :param base: string represeting base i.e. ETH
    :param base: string represeting quote i.e. BTC
    :param interval: string representing interval i.e. '5m'
    :param callable: method to recieve new candle (must take candle as a param)
    :return: none
    """
    with lock:
        topic = str(exchange_id + base + "/" + quote + interval)
        get_market_watcher(exchange_id, base, quote, interval)
        print("Subscribing to " + topic)
        pub.subscribe(callable, topic)


class BaseIndicator:
    """Base class for indicators that handles the data, keeping it in RAM so no DB calls are needed"""
    def __init__(self, market, interval, periods):
        self.market = market
        self.interval = interval
        self.periods = periods
        self.market.apply_indicator(self)

    def next_calculation(self, candle):
        pass



import os
from sqlalchemy import create_engine
from sqlalchemy import Table, Column, Integer, String, Float, MetaData, ForeignKey
from threading import Lock
import logging

logger = logging.getLogger(__name__)

db_name = 'core.db'
db_fullpath = os.path.join(
    os.path.dirname(
        os.path.realpath(__file__)),
    db_name)
lock = Lock()
engine = create_engine(
    'sqlite:///{}'.format(db_fullpath),
    connect_args={
        'check_same_thread': False},
    echo=False)
metadata = MetaData()


OHLCV = Table('OHLCV', metadata,
              Column('ID', Integer, primary_key=True),
              Column('Exchange', String),
              Column('Pair', String),
              Column('Timestamp', String),
              Column('Open', Float),
              Column('High', Float),
              Column('Low', Float),
              Column('Close', Float),
              Column('Volume', Float),
              Column('Interval', String),
              Column('TimestampRaw', Integer),
              Column('PairID', String, ForeignKey('TradingPairs.ID')),
              )

TradingPairs = Table('TradingPairs', metadata,
                     Column('ID', Integer, primary_key=True),
                     Column('Exchange', String),
                     Column('BaseCurrency', String),
                     Column('QuoteCurrency', String),
                     Column('Interval', String)
                     )


TradingOrders = Table('TradingOrders', metadata,
                      Column('Timestamp', String),
                      Column('OrderID', Integer, primary_key=True),
                      Column('Exchange', String),
                      Column('Pair', String),
                      Column('Position', String),
                      Column('Amount', Float),
                      Column('Price', Float),
                      Column('Simulated', String)
                      )


def drop_tables():
    print('Dropping tables...')
    metadata.drop_all(engine)


def create_tables():
    metadata.create_all(engine)


def reset_db():
    print('Resetting database...')
    drop_tables()
    create_tables()



from core.database import database
from sqlalchemy.sql import select, and_, func
import pandas as pd

engine = database.engine
conn = engine.connect()


class Portfolio:
    def __init__(self, market):
        """Mini portfolio object for each exchange/pair that has traded positions"""
        self.market = market
        self.exchange = self.market.exchange.id
        self.pair = self.market.exchange.analysis_pair

    def get_profit_for_trades(self):
        self.profit = get_profit_for_pair(self.exchange, self.pair)
        print("The total profit for this pair is {} up to this point in the trading session".format(self.profit))
        return self.profit

    def get_average_profit_per_trades(self):
        average_profit_per_trade = (self.profit / get_number_of_trades(self.exchange, self.pair)) / 2
        print("The average profit up to this point in the trading session is {}".format(average_profit_per_trade))
        return average_profit_per_trade


def get_profit_for_pair(exchange, pair):
    """Iterates through all trades for given exchange pair over the course of trading. Starts by subtracting the long positions (the buys) and adding the short positions (the sells) to arrive at the difference (profit"""
    """The buys are always the even rows and the sells are the odd rows (buy always before sell starting from zero)"""
    profit = 0
    counter = 0
    s = select([database.TradingPositions]).where(and_(database.TradingPositions.c.Exchange == exchange, database.TradingPositions.c.Pair == pair))
    result = conn.execute(s)

    for row in result:
        if counter % 2 == 0:
            profit = profit - row[5]
            counter += 1
        else:
            profit = profit + row[5]
            counter += 1
        return profit


def get_number_of_trades(exchange, pair):
    s = select([func.count()]).where(and_(database.TradingPositions.c.Exchange == exchange, database.TradingPositions.c.Pair == pair)).select_from(database.TradingPositions)
    result = conn.execute(s)
    return int(result)


def get_trades_for_pair_as_df(exchange, pair):
    """Returns all trades for given exchange pair over the course of trading in a dataframe"""
    s = select([database.TradingPositions]).where(and_(database.TradingPositions.c.Exchange == exchange, database.TradingPositions.c.Pair == pair))
    result = conn.execute(s)
    df = pd.DataFrame(result.fetchall())
    df.columns = result.keys()
    result.close()
    return df



from threading import Thread
from pubsub import pub
from core.markets import market
from core.markets import market_watcher
from strategies import base_strategy
import time
import logging

logger = logging.getLogger(__name__)

tickers = {}


def subscribe(tick_callable, interval):
    start_ticker(interval)
    pub.subscribe(tick_callable, "tick" + interval)


def start_ticker(interval):
    """Start a ticker/timer that notifies market watchers when to pull a new candle"""
    if interval not in tickers:
        tickers[interval] = Thread(
            target=__start_ticker, args=(
                interval,)).start()


def __start_ticker(interval):
    """Start a ticker own its own thread, will use pypubsub to send a message each time interval"""
    logger.info(interval + " ticker running...")
    live_tick_count = 0
    while True:
        """Running this 'ticker' from the main loop to trigger listeners to pull candles every 5 minutes"""
        logger.info("Live Tick: {}".format(str(live_tick_count)))
        print(interval + " tick")
        pub.sendMessage("tick" + interval)
        live_tick_count += 1
        time.sleep(__convert_interval_to_int(interval))


def __convert_interval_to_int(interval):
    if interval == "5m":
        return 300
    if interval == "1m":
        return 60
    if interval == "1h":
        return 3600



from core.database import database
from strategies import poc_strategy
from time import sleep
import logging

logger = logging.getLogger(__name__)



def start_strategy(user_input):

    # strategy = poc_strategy.PocStrategy("5m", 'bittrex', 'ETH', 'BTC', 100, 700, True, sim_balance=10)
    # strategy2 = poc_strategy.PocStrategy("5m", 'binance', 'ETH', 'BTC', 233, 900, True, sim_balance=10)

    strategy = poc_strategy.PocStrategy(user_input[3], user_input[0], user_input[1], user_input[2], user_input[4], user_input[6], user_input[5], user_input[7]) # if test strategy button clicked

    strategy.run_simulation()
    strategy.start()


def start_database():
    database.create_tables()

# For running app headless
def start():
    try:
        # wipe and recreate tables
        database.create_tables()

        start_strategy()

    except Exception as e:
        print(e)

    finally:
        database.engine.dispose()


if __name__ == "__main__":
    start()



"""Module to create, remove, and manage trades currently in play in running strategies"""
from core.markets.order import Order
import logging

logger = logging.getLogger(__name__)

positions = []


class Position:
    def __init__(self, market, amount, price):
        self.market = market
        self.amount = amount
        self.price = price

    def update(self):
        pass


class LongPosition(Position):
    """This class will handle a position's orders, stop losses, and exit/entry"""
    def __init__(self, market, amount, price, fixed_stoploss_percent, trailing_stoploss_percent, profit_target_percent):
        super().__init__(market, amount, price)
        self.is_open = False
        self.profit_target_percent = profit_target_percent
        self.trailing_stoploss_percent = self.price * trailing_stoploss_percent
        self.trailing_stoploss = self.calculate_trailing_stoploss()
        self.fixed_stoploss = price * fixed_stoploss_percent  # we can pass in an actual value to keep our fixed loss at
        self.profit_target = self.calculate_profit_target()
        self.initial_order = None

    def open(self):
        """Use the market to place the order"""
        self.initial_order = self.market.limit_buy(self.amount, self.price)
        self.is_open = True

    def update(self):
        """Use this method to trigger position to check if profit target has been met, and re-set trailiing stop loss"""
        if not self.is_open:
            pass
        elif self.market.get_best_bid() < self.trailing_stoploss or \
                self.market.get_best_bid() < self.fixed_stoploss or \
                self.market.get_best_bid() >= self.profit_target:  # check price against last calculated trailing stoploss
            self.liquidate_position()
        # re-calculate trailing stoploss
        self.trailing_stoploss = self.calculate_trailing_stoploss()

    # calculate trailing stoploss based on percent (passed in as decimal for now)
    # for example if trailing_stoploss_percent = .97
    # and latest candle low is $100
    # the trailing_stoploss will be $97
    # using low for now, but we can change this
    def calculate_trailing_stoploss(self):
        return self.price * self.trailing_stoploss_percent

    # calculate profit target based on a percent (passed in as decimal for now)
    # if buy price was $100 and profit_target_percent = 1.03
    # profit target will be $103
    def calculate_profit_target(self):
        return self.price * self.profit_target_percent

    def update_trailing_stoploss(self):
        """Will use this method to actually create the order that will serve as the stop loss"""
        pass

    def liquidate_position(self):
        """Will use this method to actually create the order that liquidates the position"""
        logger.info("Liquidating long position of " + self.amount + " | " + self.market.analysis_pair)
        self.market.limit_sell(self.amount, self.market.get_best_bid())
        self.is_open = False


class ShortPosition(Position):
    """Short position is basically just to close out the order successfully ie liquidate_position"""
    def __init__(self, market, amount, price):
        super().__init__(market, amount, price)
        self.initial_order = None

    def open(self):
        self.initial_order = Order(self.market, "sell", "limit", self.amount, self.price)

    def confirm_sell_order(self):
        pass


def open_long_position(market, amount, price, fixed_stoploss_percent, trailing_stoploss_percent, profit_target_percent):
    position = LongPosition(market, amount, price, fixed_stoploss_percent, trailing_stoploss_percent, profit_target_percent)
    position.open()
    return position


def open_short_position(market, amount, price):
    position = ShortPosition(market, amount, price).open()
    position.open()
    return position


def calculate_transaction_fee(exchange, pair):
    return exchange.load_market(pair)['fee']


def calculate_drawdown():
    pass





from pyti.exponential_moving_average import exponential_moving_average as ema
from ta.base_indicator import BaseIndicator


class ExponentialMovingAverage(BaseIndicator):
    def __init__(self, market, interval, periods):
        super().__init__(market, interval, periods)
        self.value = None

    def next_calculation(self, candle):
        """get latest candles from market, do calculation, write results to db"""
        dataset = self.market.candles[self.interval]            # save reference of market candles for read-ability
        if len(dataset) >= self.periods:                        # check that enough market candles are available for calculation
            data_window = dataset[-self.periods:]               # take slice of correct number of candles
            data = list(c[4] for c in data_window)              # take list of close values from candles
            self.value = round(ema(data, self.periods)[-1], 6)  # update current value




import ccxt
import logging
import os
from core.markets import order
from collections import defaultdict
from core.markets import market_watcher
from ccxt import BaseError


logger = logging.getLogger(__name__)

markets = []


class Market:
    """An object that allows a specific strategy to interface with an exchange,
    This includes functionality to contain and update TA indicators as well as the latest OHLCV data
    This also handles the API key for authentication, as well as methods to place orders"""

    def __init__(self, exchange, base_currency, quote_currency, strategy):
        exchange = getattr(ccxt, exchange)
        self.strategy = strategy
        self.api_key = None
        self.secret_key = None
        self.get_exchange_login()
        self.exchange = exchange({'apiKey': self.api_key, 'secret': self.secret_key, })
        self.base_currency = base_currency
        self.quote_currency = quote_currency
        self.analysis_pair = '{}/{}'.format(self.base_currency, self.quote_currency)
        self.signals = []
        self.ohlcv_id = defaultdict(int)
        self.indicators = defaultdict(list)
        self.candles = defaultdict(list)
        self.latest_candle = defaultdict(list)  # allows for order simulations based on historical ohlcv data
        markets.append(self)

    def update(self, interval, candle):
        """Notify all indicators subscribed to the interval of a new candle"""
        self.latest_candle[interval] = candle
        self.candles[interval].append(candle)
        self.do_ta_calculations(interval, candle)

    def do_ta_calculations(self, interval, candle):
        """update TA indicators applied to market"""
        for indicator in self.indicators[interval]:
            indicator.next_calculation(candle)

    def apply_indicator(self, indicator):
        """Add indicator to list of indicators listening to market's candles"""
        self.indicators[indicator.interval].append(indicator)

    def get_exchange_login(self):
        """Put API Key and Secret into login-real.txt file on your local machine"""
        try:
            login_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'login-real.txt')
            with open(login_file) as f:
                data = f.read().splitlines()
            self.exchange.api_key = data[0]
            self.exchange.secret_key = data[1]
        except:
            logger.error("Invalid login file")

    def limit_buy(self, quantity, price):
        """Place a limit buy order"""
        try:
            self.strategy.send_message("Executed buy of " + str(quantity) + " " + self.base_currency + " for " + str(price) + " " + self.quote_currency)
            return order.Order(self, "buy", "limit", quantity, price)
        except BaseError:
            self.strategy.send_message("Error creating buy order")
            logger.error("Error creating buy order")

    def limit_sell(self, quantity, price):
        """Place a limit sell order"""
        try:
            self.strategy.send_message("Executed sell of " + str(quantity) + " " + self.base_currency + " for " + str(price) + " " + self.quote_currency)
            return order.Order(self, "sell", "limit", quantity, price)
        except BaseError:
            self.strategy.send_message("Error creating sell order")
            logger.error("Error creating sell order")

    def get_wallet_balance(self):
        """Get wallet balance for quote currency"""
        try:
            logger.info(self.exchange.fetch_balance())
            return self.exchange.fetch_balance()
        except BaseError:
            logger.error("Not logged in properly")

    def get_best_bid(self):
        orderbook = self.exchange.fetch_order_book(self.analysis_pair)
        return orderbook['bids'][0][0] if len(orderbook['bids']) > 0 else None

    def get_best_ask(self):
        orderbook = self.exchange.fetch_order_book(self.analysis_pair)
        return orderbook['asks'][0][0] if len(orderbook['asks']) > 0 else None

    def get_historical_candles(self, interval, candle_limit=None):
        if len(self.candles[interval]) == 0:
            self.candles[interval] = market_watcher.get_market_watcher(self.exchange.id, self.base_currency, self.quote_currency, interval).get_historical_candles()
        if candle_limit is None:
            return self.candles[interval]
        else:
            return self.candles[interval][-candle_limit:]


from pyti.bollinger_bands import upper_bollinger_band as calc_upper_bband
from pyti.bollinger_bands import lower_bollinger_band as calc_lower_bband
from pyti.bollinger_bands import middle_bollinger_band as calc_middle_bband
from pyti.bollinger_bands import bandwidth as calc_bandwidth
from pyti.bollinger_bands import bb_range as calc_range
from ta.base_indicator import BaseIndicator


class BollingerBands(BaseIndicator):
    def __init__(self, market, interval, periods):
        super().__init__(market, interval, periods)
        self.value = None
        self.upper_band = None
        self.middle_band = None
        self.lower_band = None
        self.bandwidth = None
        self.range = None

    def next_calculation(self, candle):
        """get latest candles from market, do calculation, write results to db"""
        dataset = self.market.candles[self.interval]            # save reference of market candles for read-ability
        if len(dataset) >= self.periods:                        # check that enough market candles are available for calculation
            data_window = dataset[-self.periods:]               # take slice of correct number of candles
            data = list(c[4] for c in data_window)              # take list of close values from candles
            self.upper_band = round(calc_upper_bband(data, self.periods)[-1], 6)  # update upper band
            self.middle_band = round(calc_middle_bband(data, self.periods)[-1], 6)  # update middle band
            self.lower_band = round(calc_lower_bband(data, self.periods)[-1], 6)  # update lower band
            self.bandwidth = round(calc_bandwidth(data, self.periods)[-1], 6)  # update bandwidth
            self.range = round(calc_range(data, self.periods)[-1], 6)  # update range



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Name    : setup.py
# @Author  : yanlee

import setuptools
import shutil

# 删除dist/目录
shutil.rmtree('dist', ignore_errors=True)

setuptools.setup(
    name="Titan-pro",
    version="1.2.1",
    author="yanjlee",
    author_email="yanjlee@163.com",
    description="Titan is a totally free cryptocurrency automated trading framework for traders of all skill levels. Built on top of the ccxt API library, it works out of the box with any exchange and trading pair. We believe in decentralizing trading strategies and providing everyone access to building advanced crypto portfolios. The future of algorithmic trading is here!.",  # 模块简介
    install_requires=[
        'requests',
        'faker',
        'execjs',
        'loguru',
        'base64',
        'hashlib',
        'Crypto',
        'pandas',
        'fuzzywuzzy',
        'httpx',
        'Pillow',
        'playwright',
        'PyExecJS',
        'redis',
        'fastapi',
        'uvicorn',
        'APScheduler',
        'beautifulsoup4',
        'bs4',
        'certifi',
        'clickhouse-driver',
        'curl-cffi',
        'DrissionPage',
        'fake-useragent',
        'Flask',
        'Flask-APScheduler',
        'Flask-Cors',
        'frida',
        'gevent',
        'httpx',
        'Jinja2',
        'langchain',
        'langchain-community',
        'suiutils-py',
    ],
    long_description=open(r'readme.md', encoding='utf-8').read(),  # 读取readme自述文件
    long_description_content_type="text/markdown",
    url="https://github.com/yanjlee/Titan",  # 模块github地址
    packages=setuptools.find_packages(),     # 自动列出项目下的包
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",   # 开源许可证
        "Operating System :: OS Independent",      # 这里的定义是系统无关（全平台兼容），如果你的包只能在部分特定系统上运行，需要修改。
    ],
)


from core.database import database
from sqlalchemy.sql import select, and_
import pandas as pd
import datetime
from sqlalchemy.exc import IntegrityError
import logging


logger = logging.getLogger(__name__)
engine = database.engine
conn = engine.connect()


def insert_data_into_ohlcv_table(exchange, pair, interval, candle, pair_id):
    """Inserts exchange candle data into table"""
    logger.info('Adding candle with timestamp: ' + str(candle[0]))
    with database.lock:
        args = [exchange, pair, candle[0], candle[1], candle[2], candle[3], candle[4], candle[5], interval]
        ins = database.OHLCV.insert().values(Exchange=args[0], Pair=args[1], Timestamp=convert_timestamp_to_date(args[2]), Open=args[3], High=args[4], Low=args[5], Close=args[6], Volume=args[7], Interval=args[8], TimestampRaw=args[2], PairID=pair_id)
        conn.execute(ins)


def get_latest_candle(exchange, pair, interval):
    """Returns only latest candle if it exists, otherwise returns 0"""
    with database.lock:
        logger.info("Querying latest candle for " + exchange + " " + pair)
        s = select([database.OHLCV]).where(and_(database.OHLCV.c.Exchange == exchange, database.OHLCV.c.Pair == pair, database.OHLCV.c.Interval == interval)).order_by(database.OHLCV.c.TimestampRaw.desc()).limit(1)
        result = conn.execute(s)
        row = result.fetchone()
        result.close()
        return row

def get_all_candles(pair_id):
    with database.lock:
        logger.info("Retrieving all candles for pair " + str(pair_id))
        result = conn.execute("SELECT TimestampRaw, Open, High, Low, Close, Volume FROM OHLCV WHERE PairID = ?", (pair_id,))
        return [row for row in result]


def get_latest_N_candles_as_df(exchange, pair, interval, N):
    """Returns N latest candles for TA calculation purposes"""
    with database.lock:
        s = select([database.OHLCV]).where(and_(database.OHLCV.c.Exchange == exchange, database.OHLCV.c.Pair == pair, database.OHLCV.c.Interval == interval)).order_by(database.OHLCV.c.ID.desc()).limit(N)
        result = conn.execute(s)
        df = pd.DataFrame(result.fetchall())
        df.columns = result.keys()
        result.close()
        return df


def write_trade_pairs_to_db(exchange_id, base, quote, interval):
    """Returns the ID of the trade pair"""
    logger.info("Writing market data pair to DB")
    with database.lock:
        try:
            s = select([database.TradingPairs]).where(
                and_(database.TradingPairs.c.Exchange == exchange_id,
                     database.TradingPairs.c.BaseCurrency == base,
                     database.TradingPairs.c.QuoteCurrency == quote,
                     database.TradingPairs.c.Interval == interval))
            result = conn.execute(s).fetchone()
            if result is None:
                ins = database.TradingPairs.insert().values(Exchange=exchange_id, BaseCurrency=base, QuoteCurrency=quote, Interval=interval)
                conn.execute(ins)
                return conn.execute(s).cursor.lastrowid
            else:
                logger.info("Market data already available for pair, returning ID for lookups")
                return result[0]
        except IntegrityError:
            print("Pair already logged in DB")



def convert_timestamp_to_date(timestamp):
    value = datetime.datetime.fromtimestamp(float(str(timestamp)[:-3]))  #this might only work on bittrex candle timestamps
    return value.strftime('%Y-%m-%d %H:%M:%S')




import datetime
import time
import logging
from ccxt import OrderNotFound
from core.database import database


logger = logging.getLogger(__name__)

engine = database.engine
conn = engine.connect()


class Order:
    """Class that represents an order. Order executes on instantiation by a thread pool"""
    def __init__(self, market, side, type, amount, price):
        self.market = market
        self.side = side
        self.type = type
        self.amount = amount
        self.price = price
        self.__order_receipt = None
        logger.info("Opening " + side + " order of " + amount + " " + self.market.base_currency)
        self.execute()

    def execute(self):
        if self.type == "limit":
            if self.side == "buy":
                self.__order_receipt = self.market.exchange.create_limit_buy_order(self.market.analysis_pair, self.amount, self.price)
                write_order_to_db(self.market.exchange.id, self.market.analysis_pair, 'long', self.amount, self.price, "live")
            elif self.side == "sell":
                self.__order_receipt = self.market.exchange.create_limit_sell_order(self.market.analysis_pair, self.amount, self.price)
                write_order_to_db(self.market.exchange.id, self.market.analysis_pair, 'short', self.amount, self.price, "live")
            else:
                logger.error("Invalid order side: " + self.side + ", specify 'buy' or 'sell' ")
        elif self.type == "market":
            logger.error("Market orders not available")
        else:
            logger.error("Invalid order type: " + self.type + ", specify 'limit' or 'market' ")

    def get_id(self):
        return self.__order_receipt.get().id

    def cancel(self):
        try:
            self.market.exchange.cancel_order(self.get_id())
        except OrderNotFound:
            logger.error("Order cannot be canceled. Has already been filled")

    def is_open(self):
        return self.market.exchange.fetch_order(self.get_id())['remaining'] > 0

    def get_status(self):
        return self.market.exchange.fetch_order(self.get_id())['status']

    def get_amount_filled(self):
        return self.market.exchange.fetch_order(self.get_id())['filled']

    def get_amount_remaining(self):
        return self.market.exchange.fetch_order(self.get_id())['remaining']


def write_order_to_db(exchange, pair, position, amount, price, simulated):
    with database.lock:
        ins = database.TradingOrders.insert().values(Timestamp=get_timestamp(), Exchange=exchange, Pair=pair, Position=position, Amount=amount, Price=price, Simulated=simulated)
        conn.execute(ins)
        logger.info("Wrote open order to DB...")


def get_timestamp():
    return datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
